"""
Module for recalculate shooting position
"""
import pandas as pd
import numpy as np
from scipy.spatial.transform import Rotation as R
from borea.utils.miscellaneous.miscellaneous import angle_degree_2vect, min_max_pt, normalize
from borea.worksite.worksite import Worksite
from borea.datastruct.camera import Camera
from borea.datastruct.shot import Shot
from borea.transform_world_image.transform_shot.world_image_shot import WorldImageShot
from borea.transform_world_image.transform_shot.image_world_shot import ImageWorldShot
# pylint: disable-next=line-too-long
from borea.transform_world_image.transform_worksite.image_world_intersection import WorldIntersection  # noqa: E501
from borea.utils.miscellaneous.param_bundle import set_param_bundle_diff


class SpaceResection:
    """
    Recalculates the shot's 6 external orientation parameters,
    the 3 angles omega, phi, kappa and its position x, y, z.

    Args:
        work (Woksite): Worksite to make space resection
    """
    def __init__(self, work: Worksite) -> None:
        self.work = work

    def space_resection_to_worksite(self, pt2d: pd.DataFrame, pt3d: pd.DataFrame,
                                    pinit: dict) -> None:
        """
        Calculates the shot's 6 external orientation parameters,
        the 3 angles omega, phi, kappa and its position x, y, z.
        For all shot with a variation pixel.

        Args:
            pt2d (pd.Dataframe): Table of image coordinate point with image.
            pt3d (pd.Dataframe): Table of ground coordinate point.
            pinit (dict): Dictionnary with 3d coordinate for initialization "coor_init".
        """
        for name_shot, group in pt2d.groupby("id_shot"):
            if group.shape[0] < 3:
                print(f"Shot {name_shot} has fewer than three points,"
                      " so it cannot calculate its position and orientation.")
                continue

            dfm = group.merge(pt3d, how="inner", on="id_pt")
            obs_img = np.array([dfm["column"].to_numpy(), dfm["line"].to_numpy()])
            obs_world = np.array([dfm["x"].to_numpy(), dfm["y"].to_numpy(), dfm["z"].to_numpy()])

            pinit["kappa"] = self.init_kappa(dfm)
            self.work.shots[name_shot] = self.space_resection_shot(name_shot, obs_img,
                                                                   obs_world, pinit)
        self.work.set_param_shot(self.work.approxeucli)

        if pt3d["type"].iloc[0]:
            type_pt = "gcp2d"
        else:
            type_pt = "co_points"
        self.work.set_point_image_dataframe(pt2d, type_pt)
        self.work.set_point_world_dataframe(pt3d, type_pt)

    def space_resection_on_worksite(self, add_pixel: tuple = (0, 0)) -> None:
        """
        Recalculates the shot's 6 external orientation parameters,
        the 3 angles omega, phi, kappa and its position x, y, z.
        For all shot with a variation pixel.

        Args:
            add_pixel (tuple): Factor (column, line) added on observable point.
        """
        for key_shot, item_shot in self.work.shots.items():
            self.work.shots[key_shot] = self.space_resection_gap(item_shot, add_pixel)

    def space_resection_shot(self, name_shot: str, pt_img: np.ndarray,
                             pt_world: np.ndarray, pinit: dict) -> Shot:
        """
        Calculates the shot's 6 external orientation parameters,
        the 3 angles omega, phi, kappa and its position x, y, z.

        Args:
            name_shot (str): Name of shot to calculte externa parameters.
            pt_img (np.array): Coordonnees image of points.
            pt_world (np.array): Coordonnees world of points.
            pinit (dict): Dictionnary with 3d coordinate for initialization and kappa.

        Returns:
            Shot: Adjusted shot.
        """
        # Initialization of adjusted shot
        cam = self.work.cameras[list(self.work.cameras.keys())[0]]
        shot_adjust = Shot(name_shot, pinit["coor_init"], np.array([0, 0, pinit["kappa"]]),
                           cam.name_camera, "degree", False, "opk")
        shot_adjust.set_param_eucli_shot(self.work.approxeucli)
        z_nadir = ImageWorldShot(shot_adjust, cam).image_to_world(np.array([cam.ppax, cam.ppay]),
                                                                  self.work.type_z_shot,
                                                                  self.work.type_z_shot, False)[2]
        shot_adjust.set_z_nadir(z_nadir)

        # Calculate euclidean position
        pt_eucli = shot_adjust.projeucli.world_to_eucli(pt_world)

        # Least-square methode
        shot_adjust = self.least_square_shot(shot_adjust, pt_img, pt_eucli, pt_world)

        return shot_adjust

    def init_kappa(self, dfpt: pd.DataFrame) -> float:
        """
        Calculates a kappa angle for acquisition initialization.

        Args:
            dfpt (pd.DataFrame): Coordonnees of points in image and world.

        Returns:
            float: kappa angle.
        """
        pt_img = np.array([dfpt["column"].to_numpy(), dfpt["line"].to_numpy()])
        pt_mini, arg_min = min_max_pt(pt_img, "min")
        pt_maxi, arg_max = min_max_pt(pt_img, "max")
        pt_minw = dfpt.iloc[arg_min].get(["x", "y"]).to_numpy()
        pt_maxw = dfpt.iloc[arg_max].get(["x", "y"]).to_numpy()
        vect_i = pt_maxi - pt_mini
        vect_w = pt_maxw - pt_minw
        norm_vi = normalize(vect_i)[0]
        norm_vw = normalize(vect_w)[0]
        kappa = angle_degree_2vect(norm_vi, norm_vw)
        return np.floor(kappa)

    def space_resection_gap(self, shot: Shot, add_pixel: tuple = (0, 0)) -> Shot:
        """
        Recalculates the shot's 6 external orientation parameters,
        the 3 angles omega, phi, kappa and its position x, y, z.

        Args:
            shot (Shot): Shot to recalculte externa parameters.
            add_pixel (tuble): Pixel to be added to change marker.

        Returns:
            Shot: Adjusted shot.
        """
        # Initialization observation point
        obs, pt_world = self.take_obs(shot)

        # Calculate euclidean position
        pt_eucli = shot.projeucli.world_to_eucli(pt_world)

        # Add factor
        obs[0] += add_pixel[0]
        obs[1] += add_pixel[1]

        # Initialization of adjusted shot
        shot_adjust = Shot(shot.name_shot, shot.pos_shot, shot.ori_shot,
                           shot.name_cam, shot.unit_angle, shot.linear_alteration, shot.order_axe)
        shot_adjust.set_param_eucli_shot(shot.approxeucli)
        shot_adjust.set_z_nadir(shot.z_nadir)

        # Least-square methode
        shot_adjust = self.least_square_shot(shot_adjust, obs, pt_eucli, pt_world)

        shot_adjust.co_points = shot.co_points
        shot_adjust.gcp2d = shot.gcp2d
        shot_adjust.gcp3d = shot.gcp3d

        return shot_adjust

    def least_square_shot(self, shot_adjust: Shot, obs: np.ndarray,
                          pt_eucli: np.ndarray, pt_world: np.ndarray) -> Shot:
        """
        Least-square methode to calcule the shot's 6 external orientation parameters.

        Args:
            shot_adjust (Shot): The shot to adjust.
            obs (np.array): Observation of point in image [c_obs, l_obs].
            pt_eucli (np.array): Observation of world point in euclidean system [X, Y, Z].
            pt_world (np.array): Observation of world point [X, Y, Z].

        Returns:
            Shot: Adjusted shot.
        """
        bool_iter = True
        count_iter = 0
        while bool_iter:
            count_iter += 1

            dx = self.func_least_square(shot_adjust, obs, pt_eucli, pt_world)

            # Calculate new x = x0 + dx for position and rotation matrix
            new_pos_eucli = np.array([shot_adjust.pos_shot[0] + dx[0],
                                      shot_adjust.pos_shot[1] + dx[1],
                                      shot_adjust.pos_shot[2] + dx[2]])
            new_mat_eucli = shot_adjust.mat_rot_eucli @ R.from_rotvec(dx[3:]).as_matrix()

            # Creation of new shot with new parameter
            imc_new_adjust = Shot.from_param_euclidean(shot_adjust.name_shot, new_pos_eucli,
                                                       new_mat_eucli, shot_adjust.name_cam,
                                                       shot_adjust.unit_angle,
                                                       shot_adjust.linear_alteration,
                                                       shot_adjust.order_axe,
                                                       shot_adjust.approxeucli)
            imc_new_adjust.set_z_nadir(shot_adjust.z_nadir)

            # Look difference to know if you want to stop the calculation
            diff_coord = np.array([imc_new_adjust.pos_shot]) - np.array([shot_adjust.pos_shot])
            diff_opk = np.array([imc_new_adjust.ori_shot]) - np.array([shot_adjust.ori_shot])

            if (np.all(diff_coord < 10 ** -3) and np.all(diff_opk < 10 ** -6)) or count_iter > 100:
                # print(count_iter)
                bool_iter = False

            # Replace adjusted place
            shot_adjust = imc_new_adjust

        return shot_adjust

    def func_least_square(self, shot_adjust: Shot, obs: np.ndarray,
                          pt_eucli: np.ndarray, pt_world: np.ndarray) -> np.ndarray:
        """
        Calculate the least-squares equation dx = (A.T @ A)**-1 @ A.T @ B

        Args:
            shot_adjust (Shot): The shot to adjust.
            obs (np.array): Observation of point in image [c_obs, l_obs].
            pt_eucli (np.array): Observation of world point in euclidean system [X, Y, Z].
            pt_world (np.array): Observation of world point [X, Y, Z].

        Returns:
            np.array: dx.
        """
        cam = self.work.cameras[shot_adjust.name_cam]
        # Calculate position column and line with new shot f(x0)
        f0 = WorldImageShot(shot_adjust,
                            cam
                            ).world_to_image(pt_world,
                                             self.work.type_z_data,
                                             self.work.type_z_shot)

        # Calculate residual vector B
        v_res = np.c_[obs[0] - f0[0], obs[1] - f0[1]].reshape(2 * len(pt_eucli[0]))

        # Creation of A with mat_obs_axia
        # Calculate dx = (A.T @ A)**-1 @ A.T @ B
        return np.squeeze(np.linalg.lstsq(self.mat_obs_axia(pt_eucli, shot_adjust),
                                          v_res, rcond=None)[0])

    def mat_obs_axia(self, pt_eucli: np.ndarray, imc_adjust: Shot) -> np.ndarray:
        """
        Setting up the mat_a matrix to solve the system by axiator.

        Args:
            pt_eucli (np.array): Coordinate [X, Y, Z] euclidean.
            imc_adjust (Shot): adjusted shot.

        Returns:
            np.array: Matrix A.
        """
        vect_a, vect_u, mat_v = set_param_bundle_diff(imc_adjust, pt_eucli)

        # Axiator of vect_a
        a_axiator = np.zeros((3 * len(vect_a[0]), 3))
        a_axiator[0::3, 1] = -vect_a[2]
        a_axiator[0::3, 2] = vect_a[1]
        a_axiator[1::3, 0] = vect_a[2]
        a_axiator[1::3, 2] = -vect_a[0]
        a_axiator[2::3, 0] = -vect_a[1]
        a_axiator[2::3, 1] = vect_a[0]

        cam = self.work.cameras[imc_adjust.name_cam]
        mat_a = -np.tile(np.repeat(cam.focal / vect_u[2] ** 2, 2), (6, 1)).T
        mat_a[:, :3] *= (mat_v @ imc_adjust.mat_rot_eucli)

        mat_a[:, 3:] *= np.einsum('lij, ljk->lik',
                                  (mat_v @ imc_adjust.mat_rot_eucli).reshape(-1, 2, 3),
                                  # pylint: disable-next=too-many-function-args
                                  a_axiator.reshape(-1, 3, 3)).reshape(-1, 3)

        return mat_a

    def take_obs(self, shot: Shot) -> tuple:
        """
        Check co point on the shot to use in observation.
        If there aren't enough points, add 20 random observation points.

        Args:
            shot (Shot): The shot to adjust.

        Returns:
            tuple: np.array(obs_image), np.array(pt_world).
        """
        add_pt = True
        pt_world_random = 0
        if shot.co_points:
            if not self.work.co_pts_world:
                WorldIntersection(self.work).calculate_image_world_by_intersection("co_points")
            obs, pt_world = self.work.get_coor_pt_img_and_world(shot.name_shot, "co_points")
            if obs.shape[1] >= 7:
                add_pt = False

        if add_pt:
            cam = self.work.cameras[shot.name_cam]
            # Initialization of 20 points for shooting position
            obs_random, z_world = self.seed_20_point(cam)
            # Calculate world position
            x_world, y_world, _ = ImageWorldShot(shot,
                                                 cam
                                                 ).image_z_to_world(obs_random,
                                                                    self.work.type_z_shot,
                                                                    z_world)
            pt_world_random = np.array([x_world, y_world, z_world])

        if add_pt and shot.co_points:
            obs = np.concatenate((obs, obs_random), axis=1)
            pt_world[2] += 350  # 350 is the mean of z_obs in seed_20_point
            pt_world = np.concatenate((pt_world, pt_world_random), axis=1)
            add_pt = False

        if add_pt:
            obs = obs_random
            pt_world = pt_world_random

        return obs, pt_world

    def seed_20_point(self, cam: Camera) -> tuple:
        """
        Positioning of 20 points on an image by percentage position on width and height.
        The z-world position is given by fixed values.

        Args:
            cam (Camera): Camera of the shot.

        Returns:
            np.array: Tuple of 3 elements (position column obs, position line obs, z world).
        """
        pourcent_x = np.array([6.25, 12.5, 18.75, 25, 31.25, 37.5, 37.5, 43.75, 50, 56.25,
                               62.5, 62.5, 68.75, 68.75, 75, 81.25, 87.5, 87.5, 93.75, 93.75]) / 100
        pourcent_y = np.array([80, 20, 60, 90, 30, 10, 50, 70, 20, 40,
                               60, 90, 20, 80, 50, 20, 10, 70, 40, 90]) / 100
        z_obs = np.array([200, 320, 250, 240, 330, 335, 340, 330, 350, 360,
                          360, 350, 370, 355, 380, 400, 450, 400, 500, 300])
        c_obs = pourcent_x * cam.width
        l_obs = pourcent_y * cam.height

        return np.array([c_obs, l_obs]), z_obs



"""
Script to read camera file txt or xml.
"""
from pathlib import Path, PureWindowsPath
from borea.datastruct.camera import Camera
from borea.worksite.worksite import Worksite


def read_camera(files: list, work: Worksite) -> None:
    """
    Manage file in list files to read.

    Args:
        files (list): Path list of files cameras.
        work (Worksite): Worksite which needs camera data.
    """
    for file in files:
        camera_txt(Path(PureWindowsPath(file)), work)


def camera_txt(file: Path, work: Worksite) -> None:
    """
    Read txt file camera.

    Args:
        files (Path): Path list of files cameras.
        work (Worksite): Worksite which needs camera data.
    """
    dict_info = {}
    try:
        with open(file, 'r', encoding="utf-8") as file_cam:
            for info_cam in file_cam.readlines():
                if info_cam != '\n' and info_cam[0] != '#':
                    # Camera info retrieval
                    info_name, info_data = info_cam.split(" = ")
                    dict_info[info_name.lower()] = info_data[:-1]

            file_cam.close()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The path {file} is incorrect !!!") from e

    # Create Camera
    cam = Camera(dict_info["name"])

    for name_attr, item in dict_info.items():
        if name_attr != "name":
            type_item = int if name_attr in ["width", "height"] else float
            try:
                setattr(cam, name_attr.lower(), type_item(item))
            except ValueError:
                continue
    work.cameras[dict_info["name"]] = cam



"""
Args of parser for unit of shot
"""
import argparse


def args_input_shot(parser: argparse) -> argparse:
    """
    Args for reading opk file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('-b', '--order_axe_input',
                        type=str, default="opk",
                        help="Order of rotation matrix axes.")
    parser.add_argument('-u', '--unit_angle',
                        type=str, default="degree", choices=['degree', 'radian'],
                        help="Unit of the angle of shooting, 'degree' or 'radian'.")
    parser.add_argument('-a', '--linear_alteration',
                        type=bool, default=True,
                        help="True if z shot corrected by linear alteration.")
    return parser


def args_output_shot(parser: argparse) -> argparse:
    """
    Args for reading opk file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    parser.add_argument('--ob', '--order_axe_output',
                        type=str, default=None,
                        help="Order of rotation matrix axes you want in output.")
    parser.add_argument('--ou', '--output_unit_angle',
                        type=str, default=None, choices=["degree", "radian", None],
                        help="Unit of the angle of shooting, 'degree' or 'radian'")
    parser.add_argument('--oa', '--output_linear_alteration',
                        type=bool, default=None,
                        help="True if z shot corrected by linear alteration.")
    return parser



"""
Class to calcule world coordinate by intersection.
"""
from dataclasses import dataclass
import numpy as np
from borea.worksite.worksite import Worksite
from borea.datastruct.shot import Shot
from borea.geodesy.local_euclidean_proj import LocalEuclideanProj
from borea.geodesy.approx_euclidean_proj import ApproxEuclideanProj
from borea.transform_world_image.transform_shot.conversion_coor_shot import conv_z_shot_to_z_data
from borea.transform_world_image.transform_shot.image_world_shot import ImageWorldShot


@dataclass
class WorldIntersection:
    """
    Class to calculate image coordinate to world coordinate in worksite by intersect bundle.

    Args:
        name (str): Name of the worksite.
    """
    work: Worksite

    def calculate_image_world_by_intersection(self, type_point: str,
                                              control_type: list = None) -> None:
        """
        Calculates the ground position of connecting point by intersection with
        the most distance between two shots or ground image point.

        Args:
            type_point (str): "co_points" or "gcp2d"
                              depending on what you want to calculate.
            control_type (list): Type controle for gcp.
        """

        out_pt, control_type = self.work.get_attr_transfo_pt(type_point, control_type)

        for name_pt, list_shot in getattr(self.work, type_point).items():  # Loop on points
            try:
                if control_type != [] and self.work.gcp3d[name_pt].code not in control_type:
                    continue
            except KeyError:
                continue
            if len(list_shot) == 1:
                continue

            coor = self.comput_inter_in_2_more_distant_shot(name_pt, list_shot)

            getattr(self.work, out_pt)[name_pt] = coor

    def comput_inter_in_2_more_distant_shot(self, name_pt: str, list_shot: list) -> np.ndarray:
        """
        Search for the two most distant images where the point is visible,
        to calculate the point's position.

        Args:
            name_pt (str): Name of the point.
            list_shot (list): List of name shot where the point is visible.

        Returns:
            np.ndarray: World coordinate of the point.
        """
        shot1 = ""
        shot2 = ""
        dist = 0
        list_shot1 = list_shot.copy()
        list_shot2 = list_shot1.copy()
        _ = list_shot1.pop(-1)
        for name_shot1 in list_shot1:  # Double loop on shots where see the point
            _ = list_shot2.pop(0)
            for name_shot2 in list_shot2:
                pos_shot1 = self.work.shots[name_shot1].pos_shot
                pos_shot2 = self.work.shots[name_shot2].pos_shot
                new_dist = np.sqrt(np.sum((pos_shot1 - pos_shot2)**2))
                if new_dist > dist:
                    dist = new_dist
                    shot1 = name_shot1
                    shot2 = name_shot2
        return self.intersection_pt_in_2shot(name_pt,
                                             self.work.shots[shot1],
                                             self.work.shots[shot2])

    def intersection_pt_in_2shot(self, name_point: str, shot1: Shot, shot2: Shot) -> np.ndarray:
        """
        Calculates the euclidien position of a point from two shots.

        Args:
            name_point (str): Name of copoint to calcule coordinate.
            shot1 (Shot): Frist shot.
            shot2 (Shot): Second shot.

        Returns:
            np.array: Euclidien coordinate of the copoint.
        """
        # Retrieve coordinates of points in the image.
        if name_point in list(shot1.co_points):
            p_img1 = shot1.co_points[name_point]
            p_img2 = shot2.co_points[name_point]
        else:
            p_img1 = shot1.gcp2d[name_point]
            p_img2 = shot2.gcp2d[name_point]

        # Setting up a Euclidean projection centered on the two images.
        bary = (shot1.pos_shot + shot2.pos_shot)/2
        if shot1.approxeucli:
            projeucli = ApproxEuclideanProj(bary[0], bary[1])
        else:
            projeucli = LocalEuclideanProj(bary[0], bary[1])

        # Calculates data specific to Euclidean projection.
        mat_eucli1 = projeucli.mat_to_mat_eucli(shot1.pos_shot[0], shot1.pos_shot[1], shot1.mat_rot)
        mat_eucli2 = projeucli.mat_to_mat_eucli(shot2.pos_shot[0], shot2.pos_shot[1], shot2.mat_rot)
        pos_eucli1 = conv_z_shot_to_z_data(shot1, self.work.type_z_shot, self.work.type_z_data)
        pos_eucli2 = conv_z_shot_to_z_data(shot2, self.work.type_z_shot, self.work.type_z_data)
        pos_eucli1 = projeucli.world_to_eucli(pos_eucli1)
        pos_eucli2 = projeucli.world_to_eucli(pos_eucli2)

        # Calculates the director vectors of the point bundles in the Euclidean reference system.
        vect1 = mat_eucli1.T @ ImageWorldShot(shot1, self.work.cameras[shot1.name_cam]
                                              ).image_to_bundle(p_img1)
        vect2 = mat_eucli2.T @ ImageWorldShot(shot2, self.work.cameras[shot2.name_cam]
                                              ).image_to_bundle(p_img2)

        # Calculating the intersection of two lines
        pt_inter = self.intersection_line_3d(vect1, pos_eucli1, vect2, pos_eucli2)

        # Converting the point to the world system.
        pt_inter = projeucli.eucli_to_world(pt_inter)
        return pt_inter

    def intersection_line_3d(self, vect1: np.ndarray, point1: np.ndarray,
                             vect2: np.ndarray, point2: np.ndarray) -> np.ndarray:
        """
        Calculation of the intersection point between 2 line in a 3d system.

        Args:
            vect1 (np.array): Directing vector of the first line.
            point1 (np.array): A point on the first line.
            vect2 (np.array): Directing vector of the second line.
            point2 (np.array): A point on the second line.

        Returns:
            np.array: The point of the intersection of the lines.
        """
        base = point2 - point1
        norme_v1 = vect1 @ vect1
        norme_v2 = vect2 @ vect2
        v1_v2 = vect1 @ vect2
        b_v1 = base @ vect1
        b_v2 = base @ vect2
        denum = v1_v2**2 - norme_v1*norme_v2
        p1_eucli = point1 + ((b_v2*v1_v2 - b_v1*norme_v2)/(denum))*vect1
        p2_eucli = point2 + ((b_v2*norme_v1 - b_v1*v1_v2)/(denum))*vect2
        return (p1_eucli + p2_eucli) / 2



"""
Module for manipulating a cartographic system.
"""
from typing import Union, List
from dataclasses import dataclass
import pyproj
import numpy as np
from borea.geodesy.transform_geodesy import TransformGeodesy
from borea.utils.singleton.singleton import Singleton


# pylint: disable=unpacking-non-sequence
@dataclass
class ProjEngine(metaclass=Singleton):
    """
    This class provides functions for using a cartographic system.
    """
    epsg: int = None
    geoid: list = None
    epsg_output: int = None

    def __post_init__(self) -> None:
        if self.epsg:
            self.crs = pyproj.CRS.from_epsg(self.epsg[0])
            self.proj = pyproj.Proj(self.crs)
            self.tf = TransformGeodesy(self.epsg, self.geoid, self.epsg_output)

    def set_epsg(self, epsg: list, geoid: list = None, epsg_output: int = None) -> None:
        """
        Setter of the class ProjEngine.
        Allows to init the class with data.

        Args:
            epsg (list): Code epsg of the projection ex: [2154].
            geoid (list): List of geoid to use.
            epsg_output (int): Code epsg of the output projection. If you want to change.
        """
        self.epsg = epsg
        self.geoid = geoid
        self.epsg_output = epsg_output
        self.__post_init__()

    def get_meridian_convergence(self, x_carto: Union[np.ndarray, List[float], float],
                                 y_carto: Union[np.ndarray, List[float], float]) -> np.ndarray:
        """
        Compute meridian convergence.
        Values are extracted from pyproj.

        Args:
            x_carto (np.array, List[float], float): x cartographic coordinates.
            y_carto (np.array, List[float], float): y cartographic coordinates.

        Returns:
            np.array : Meridian convergence in degree.
        """
        (x_geog, y_geog) = self.tf.carto_to_geog(x_carto, y_carto)
        return -np.array(self.proj.get_factors(x_geog, y_geog).meridian_convergence)

    def get_scale_factor(self, x_carto: Union[np.ndarray, List[float], float],
                         y_carto: Union[np.ndarray, List[float], float]) -> np.ndarray:
        """
        Compute scale factor.
        Values are extracted from pyproj.

        Args:
            x_carto (Union[np.array, List[float], float]): x cartographic coordinates.
            y_carto (Union[np.array, List[float], float]): y cartographic coordinates.

        Returns:
            np.array: Scale factor and meridian convergence.
        """
        x_geog, y_geog = self.tf.carto_to_geog(x_carto, y_carto)
        return np.array(self.proj.get_factors(x_geog, y_geog).meridional_scale) - 1



"""
Function to normalize data.
"""
import numpy as np


def normalize(array: np.ndarray) -> tuple:
    """
    Normalize array.

    Args:
        array (np.array): Array data to normalize.

    Returns:
        tuple: data normalize, offset of data and scale of data.
    """
    offset = (np.min(array) + np.max(array))/2
    scale = (np.max(array) - np.min(array))/2
    array_norm = (array - offset) / scale
    return array_norm, offset, scale


def dist_2pts(pt1: np.ndarray, pt2: np.ndarray) -> float:
    """
    Calculate the distance of two points.

    Args:
        pt1 (np.array): Coordinates of the first point.
        pt2 (np.array): Coordinates of the second point.

    Returns:
        float: The distance.
    """
    return np.linalg.norm(pt1 - pt2)


def min_max_pt(array: np.ndarray, type_m: str) -> tuple:
    """
    Get min or max of coordinnate 2D in an array of dim (2,n).

    Args:
        array (np.array): Array data to get min or max.
        type_m (str): Type min or max.

    Returns:
        tuple: min or max Coordinate of one point and index in array.
    """
    if type_m == "min":
        func = np.argmin
    else:
        func = np.argmax

    arg_col = func(array[0])
    arg_lin = func(array[1])
    pt_col = array[:, arg_col]
    pt_lin = array[:, arg_lin]

    if (pt_col == pt_lin).all():
        return pt_col, arg_col

    d1 = dist_2pts(np.array([0, 0]), pt_col)
    d2 = dist_2pts(np.array([0, 0]), pt_lin)
    if (d1 <= d2 and type_m == "min") or (d1 >= d2 and type_m == "max"):
        return pt_col, arg_col

    return pt_lin, arg_lin


def angle_degree_2vect(u: np.ndarray, v: np.ndarray) -> float:
    """
    Calculates degree angle between two vectors.

    Args:
        u (np.array): Fisrt vector.
        v (np.array): Second vector.

    Returns:
        float: Degree angle between two vectors.
    """
    nu = np.linalg.norm(u)
    nv = np.linalg.norm(v)
    costheta = (u @ v)/(nu * nv)
    return np.arccos(costheta)*180/np.pi



"""
Module math for bundle parameter.
"""
import numpy as np
from borea.datastruct.shot import Shot


def set_param_bundle_diff(shot: Shot, coor_eucli: np.ndarray) -> tuple:
    """
    Setting up the A vector, U vector and V matrix parameters for the differential function
    of beam parameters, in the Euclidean reference frame.
    A = M - S
    U = R @ A
    U = [u1, u2, u3]
    V = [[u3, 0,-u1][ 0,u3,-u2]]
    M euclidean point, S euclidean position shot, R rotation matrix euclidean to image.

    Args:
        shot (Shot): The shot in which we work.
        coor_eucli (np.array): Euclidean coordinates of the point.

    Returns:
        tuple: vector A, vector U and matrix V.
    """
    vect_a = np.vstack([coor_eucli[0] - shot.pos_shot_eucli[0],
                        coor_eucli[1] - shot.pos_shot_eucli[1],
                        coor_eucli[2] - shot.pos_shot_eucli[2]])  # vect_a = M-S
    vect_u = shot.mat_rot_eucli @ vect_a  # U = RA

    mat_v = np.zeros((2 * len(vect_u[0]), 3))
    mat_v[::2, 0] = vect_u[2]
    mat_v[::2, 2] = -vect_u[0]
    mat_v[1::2, 1] = vect_u[2]
    mat_v[1::2, 2] = -vect_u[1]

    return vect_a, vect_u, mat_v



"""
A script to read opk file.
"""
import platform
from pathlib import Path
import numpy as np
from borea.worksite.worksite import Worksite
from borea.utils.check.check_args_opk import check_args_opk


def read(file: Path, args: dict, work: Worksite) -> Worksite:
    """
    Reads an opk file to transform it into a Workside object.

    Args:
        file (Path): Path to the worksite.
        args (dict): Information for reading an opk file.
                     keys:
                     "order_axe" (str): Order of rotation matrix axes,
                     "interval" (list): Interval of lines taken into account,
                     [i, j] if i or j is None = :. e.g. [1, None] = [1:].
                     "header" (list): List of column type file.
                     "unit_angle" (str): Unit of angle 'degrees' or 'radian'.
                     "linear_alteration" (bool): True if data corrected by linear alteration.
        work (Worksite): Worksite to add shot.

    Returns:
        Worksite: The worksite.
    """
    args, header, type_z = check_args_opk(args)

    try:
        with open(file, 'r', encoding="utf-8") as file_opk:
            for item_opk in file_opk.readlines()[args["interval"][0]:args["interval"][1]]:
                if item_opk != '\n' and item_opk[0] != '#':
                    item_shot = item_opk.split()
                    if len(item_shot) != len(header):
                        raise ValueError(f"The number of columns in your file {len(item_shot)}"
                                         " is different from the number of columns in your input"
                                         f" format {len(header)}.")
                    work.add_shot(item_shot[header.index("N")],
                                  np.array([float(item_shot[header.index("X")]),
                                            float(item_shot[header.index("Y")]),
                                            float(item_shot[header.index("Z")])], dtype=float),
                                  np.array([float(item_shot[header.index("O")]),
                                            float(item_shot[header.index("P")]),
                                            float(item_shot[header.index("K")])], dtype=float),
                                  item_shot[header.index("C")],
                                  args["unit_angle"], args["linear_alteration"], args["order_axe"])
            file_opk.close()
    except FileNotFoundError as e:
        raise FileNotFoundError(f"The path {file} is incorrect !!! "
                                f"or your os is {platform.system()}. "
                                "For Windows path is \\, "
                                "for Linux and MacOS (Darwin) is / .") from e

    work.type_z_shot = type_z
    return work



"""
World image transformation module for Shot
"""
import numpy as np
from borea.datastruct.shot import Shot
from borea.datastruct.camera import Camera
from borea.datastruct.dtm import Dtm
from borea.geodesy.proj_engine import ProjEngine
from borea.transform_world_image.transform_shot.conversion_coor_shot import conv_z_shot_to_z_data


class WorldImageShot():
    """
    Function world_to_image for shot.

    Args:
        shot (Shot): The shot for convert coordinate
        cam (Camera): The camera of the shot.
    """
    def __init__(self, shot: Shot, cam: Camera) -> None:
        self.shot = shot
        self.cam = cam

    def world_to_image(self, coor_world: np.ndarray,
                       type_z_data: str, type_z_shot: str) -> np.ndarray:
        """
        Calculates the c,l coordinates of a terrain point in an image.

        Args:
            coor_world (np.array): The coordinate [x, y, z] of ground point.
            type_z_data (str): Type of z of data, "height" or "altitude".
            type_z_shot (str): Type of z of worksite, "height" or "altitude".

        Returns:
            np.array: The image coordinate [c,l].
        """
        if type_z_data != type_z_shot and not ProjEngine().tf.geog_to_geoid:
            raise ValueError("Missing geoid")

        if self.shot.linear_alteration and not Dtm().path_dtm and not self.shot.approxeucli:
            raise ValueError("Missing dtm")

        if self.shot.approxeucli and not self.shot.linear_alteration:
            raise ValueError("The data are not corrected by linear alteration "
                             "and you are using an approximate system.")

        p_eucli = self.shot.projeucli.world_to_eucli(coor_world)

        # Convert coordinate in bundle system to image system
        coor_image = self.eucli_to_image(p_eucli, type_z_data, type_z_shot)

        return coor_image

    def eucli_to_image(self, p_eucli: np.ndarray,
                       type_z_data: str, type_z_shot: str) -> np.ndarray:
        """
        Convert euclidean coordinate to image coordinate.

        Args:
            p_eucli (np.array): The euclidean coordinate [x, y, z] of ground point.
            type_z_data (str): Type of z of data, "height" or "altitude".
            type_z_shot (str): Type of z of worksite, "height" or "altitude".

        Returns:
            np.array: Bundle coordinate [c,l].
        """
        # Convert euclidean coordinate system to bundle system
        p_bundle = self.eucli_to_bundle(p_eucli, type_z_data, type_z_shot)

        # Convert coordinate in bundle system to image system
        x_col, y_line = self.bundle_to_image(p_bundle)

        return np.array([x_col, y_line])

    def eucli_to_bundle(self, p_eucli: np.ndarray,
                        type_z_data: str, type_z_shot: str) -> np.ndarray:
        """
        Convert euclidean coordinate to bundle coordinate.

        Args:
            p_eucli (np.array): The euclidean coordinate [x, y, z] of ground point.
            type_z_data (str): Type of z of data, "height" or "altitude".
            type_z_shot (str): Type of z of worksite, "height" or "altitude".

        Returns:
            np.array: Bundle coordinate [c,l].
        """
        pos_shot_new_z = conv_z_shot_to_z_data(self.shot, type_z_shot, type_z_data,
                                               approx=self.shot.approxeucli)

        # Convert coordinate in world system to euclidean system
        pos_eucli = self.shot.projeucli.world_to_eucli(pos_shot_new_z)

        # Convert coordinate in euclidean system to bundle system
        p_bundle = np.squeeze((self.shot.mat_rot_eucli @ np.vstack([p_eucli[0] - pos_eucli[0],
                                                                    p_eucli[1] - pos_eucli[1],
                                                                    p_eucli[2] - pos_eucli[2]])))
        return p_bundle

    def bundle_to_image(self, p_bundle: np.ndarray) -> np.ndarray:
        """
        Convert coordinate bundle to coordinate image col line.

        Args:
            p_bundle (np.array): [X, Y, Z] coordinates in bundle.

        Returns:
            np.array: Image coordinate x_col y_line.
        """
        x_shot = p_bundle[0] * self.cam.focal / p_bundle[2]
        y_shot = p_bundle[1] * self.cam.focal / p_bundle[2]
        z_shot = p_bundle[2]
        x_shot, y_shot, z_shot = self.shot.f_sys(x_shot, y_shot, z_shot)
        x_col = self.cam.ppax + x_shot
        y_line = self.cam.ppay + y_shot

        return np.array([x_col, y_line])



"""
Photogrammetry site file reader module.
"""
import importlib
from pathlib import Path, PureWindowsPath
from borea.worksite.worksite import Worksite


def reader_orientation(file: str, args: dict) -> Worksite:
    """
    Photogrammetric site file reading function.

    Args:
        file (str): Path to the worksite.
        args (dict): Other information for reading the file.

    Returns:
        Worksite: The worksite.
    """
    # Attention multiple file management orientation
    # Attention management of files with the same extension but different formats
    file = Path(PureWindowsPath(file))
    name_work = file.stem
    ext = file.suffix[1:]

    work = Worksite(name_work)
    try:
        my_module = importlib.import_module("borea.reader.orientation.reader_" + ext.lower())
        work = my_module.read(file, args, work)
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(f"{ext} file is not taken into account !!!") from e

    return work



"""
Args of parser for writing opk file
"""
import argparse
from borea.process.p_add_data.p_unit_shot import args_output_shot
from borea.process.p_add_data.p_write import args_writer
from borea.worksite.worksite import Worksite
from borea.writer.manage_writer import manager_writer


def args_writing_opk(parser: argparse) -> argparse:
    """
    Args for writing opk file.

    Args:
        parser (argparse): Parser to add argument.

    Returns:
        argsparse: Parser with argument.
    """
    # pylint: disable=duplicate-code
    parser = args_writer(parser)
    parser.add_argument('-o', '--output_header',
                        type=str, default=None,
                        help='Type of each column in the site file.'
                        'e.g. NXYZOPKC'
                        'N: name of shot'
                        'X: coordinate x of the shot position'
                        'Y: coordinate y of the shot position'
                        'Z: coordinate z of the shot position in altitude'
                        'H: coordinate z of the shot position in height'
                        'O: omega rotation angle'
                        'P: phi rotation angle'
                        'K: kappa rotation angle'
                        'C: name of the camera')
    parser = args_output_shot(parser)
    return parser


def process_args_write_opk(args: argparse, work: Worksite) -> None:
    """
    Processing args with data.

    Args:
        args (argparse): Arg to apply on worksite (data).
        work (Worksite): Data.
    """
    if args.ob:
        args.ob = args.ob.lower()

    if args.output_header:
        args.output_header = list(args.output_header.upper())

    # Writing data
    print("Writing OPK.")
    if args.namereturn is not None:
        args_writing = {"order_axe": args.ob,
                        "header": args.output_header,
                        "unit_angle": args.ou,
                        "linear_alteration": args.oa}
        manager_writer("opk", args.namereturn, args.pathreturn, args_writing, work)
        print(f"File written in {args.pathreturn + args.namereturn}.opk.")
    else:
        raise ValueError("The name of the saving file is missing -n.")



"""
A script for verification args to read a point file.
"""
from borea.utils.check.check_header import check_head, check_h_z, get_type_z_and_header


def check_header_file(header: list, type_pt: str) -> tuple:
    """
    Check if the header of the file is good.

    Args:
        header (list): List of column type file.
        type_pt (str): Type of point is reading (co_point, gcp2d, gcp3d).

    Returns:
        tuple: header, type_z
    """
    type_z = None
    list_letter = ['S', 'N', 'X', 'Y', 'Z', 'H', 'P', 'T']

    bad_head, ms_error_letter, _, symbol = check_head(header, list_letter)

    if type_pt in ["gcp3d", "pt3d"]:
        ll_type = ['P', 'T', 'X', 'Y', 'Z', 'H']
        misss = set(ll_type) - symbol
        if misss != set() and misss != set("T"):
            if "T" in misss:
                misss.remove("T")
            bad_head, ms_error_letter = check_h_z(bad_head, misss, ms_error_letter)
        if not bad_head:
            header, type_z = get_type_z_and_header(header)
    else:
        ll_type = ['P', 'N', 'X', 'Y']
        misss = set(ll_type) - symbol
        if misss != set():
            bad_head = True
            ms_error_letter += f"The letters {misss} are missing.\n"

    ms_error = "Your header is not correct.\n"
    ms_error += ms_error_letter
    if bad_head:
        raise ValueError(ms_error)

    return header, type_z



"""
Module to conversion z for WorldImageShot and ImageWorldShot
"""
import numpy as np
from borea.datastruct.shot import Shot
from borea.geodesy.proj_engine import ProjEngine


def conv_z_shot_to_z_data(shot: Shot, type_z_shot: str, type_z_data: str,
                          nonadir: bool = True, approx: bool = False) -> np.ndarray:
    """
    Convert type z shot to type z data. The Z of the object is NOT modified.

    Args:
        shot (Shot): Shot to process.
        type_z_shot (str): Shot type z "altitude" or "height".
        type_z_data (str): Data type z "altitude" or "height".
        nonadir (bool): To calculate nadir no take linear alteration.

    Returns:
        np.array: Shot position coordinate with type z asking.
    """
    pos_shot = shot.pos_shot.copy()

    if nonadir and shot.linear_alteration and not approx:
        new_z = shot.get_z_remove_scale_factor()
        pos_shot[2] = new_z

    if type_z_shot != type_z_data:
        if type_z_shot == "height":
            new_z = ProjEngine().tf.tranform_altitude(pos_shot)
        else:
            new_z = ProjEngine().tf.tranform_height(pos_shot)
        pos_shot[2] = new_z

    return pos_shot


def conv_output_z_type(coor: np.ndarray, type_z_input: str, type_z_output: str) -> np.ndarray:
    """
    Convert type z to the output given.

    Args:
        coor (Union[float, np.ndarray]): Coordinate [X, Y, Z].
        type_z_input (str): Z type in input "height" or "altitude".
        type_z_output (str): Z type in output "height" or "altitude".

    Returns:
        np.array: Coodinate x, y, z.
    """
    new_z = coor[2]
    if type_z_input != type_z_output:
        if type_z_input == "height":
            new_z = ProjEngine().tf.tranform_altitude(coor)
        else:
            new_z = ProjEngine().tf.tranform_height(coor)

    return np.array([coor[0], coor[1], new_z])



"""
Class to write shot in conical IGN format
"""
from datetime import datetime
from pathlib import Path
import xml.etree.ElementTree as ET
from dataclasses import dataclass

import numpy as np
from borea.datastruct.camera import Camera
from borea.datastruct.shot import Shot
from borea.utils.xml.xml import format_xml, indent, add_elem

B_TIME = ["year", "month", "day", "hour", "minute", "second", "time_system"]
B_PT3D = ["x", "y", "z"]
B_ORI = ["easting", "northing", "altitude"]
B_TCAM = ["width", "height"]
B_CAM = ["c", "l", "focale"]


@dataclass
class Conl:
    """
    Class for light conical shot.
    File reading by software Geoview IGN on MacOs.

    Args:
        shot (Shot): Shot to convert in conical file.
        cam (Camera): Camera of the shot.
        proj (str): Projection of the shot for GEOVIEW.
    """
    shot: Shot
    cam: Camera
    proj: str

    def save_conl(self, path_conical: Path) -> None:
        """
        Save the shot as light conical file.

        Args:
            path_conical (Path): path to the light conical file.
        """
        date_now = datetime.now()

        # creation XML
        ori = ET.Element("orientation", {})
        ET.SubElement(ori, "lastmodificationbylibori",
                      {'date': date_now.strftime("%Y-%m-%d"),
                       "time": date_now.strftime("%H h %M min %S sec")})
        ET.SubElement(ori, "version").text = format_xml("1.0")

        auxiliary_data = ET.SubElement(ori, "auxiliarydata")
        ET.SubElement(auxiliary_data, "image_name").text = format_xml(self.shot.name_shot)

        image_date = ET.SubElement(auxiliary_data, "image_date")
        add_elem(image_date, B_TIME, [0, 0, 0, 0, 0, 0, None])

        ET.SubElement(auxiliary_data, "samples")

        self.set_geometry_xml(ori)

        tree = ET.ElementTree(ori)
        indent(ori)
        tree.write(path_conical, encoding='utf-8', xml_declaration=True)

    def set_geometry_xml(self, ori: ET) -> None:
        """
        Setup balise geometry of xml.

        Args:
            ori (ET): The xml.
        """
        pos_shot = np.round(np.copy(self.shot.pos_shot), 5)

        geometry = ET.SubElement(ori, "geometry", {'type': "physique"})

        extrinseque = ET.SubElement(geometry, "extrinseque")

        systeme = ET.SubElement(extrinseque, "systeme")

        euclidien = ET.SubElement(systeme, "euclidien", {'type': "MATISRTL"})
        add_elem(euclidien, B_PT3D[:2], pos_shot[:2])

        ET.SubElement(systeme, "geodesique").text = format_xml(self.proj).upper()

        ET.SubElement(extrinseque, "grid_alti").text = format_xml("UNKNOWN")

        sommet = ET.SubElement(extrinseque, "sommet")
        add_elem(sommet, B_ORI, [0, 0, pos_shot[2]])

        rotation = ET.SubElement(extrinseque, "rotation")
        ET.SubElement(rotation, "Image2Ground").text = format_xml('false')

        mat3d = ET.SubElement(rotation, "mat3d")

        l1_pt3d = ET.SubElement(ET.SubElement(mat3d, "l1"), "pt3d")
        add_elem(l1_pt3d, B_PT3D, [self.shot.mat_rot[0][0],
                                   self.shot.mat_rot[0][1],
                                   self.shot.mat_rot[0][2]], ".16f")

        l2_pt3d = ET.SubElement(ET.SubElement(mat3d, "l2"), "pt3d")
        add_elem(l2_pt3d, B_PT3D, [self.shot.mat_rot[1][0],
                                   self.shot.mat_rot[1][1],
                                   self.shot.mat_rot[1][2]], ".16f")

        l3_pt3d = ET.SubElement(ET.SubElement(mat3d, "l3"), "pt3d")
        add_elem(l3_pt3d, B_PT3D, [self.shot.mat_rot[2][0],
                                   self.shot.mat_rot[2][1],
                                   self.shot.mat_rot[2][2]], ".16f")

        self.set_camera_xml(geometry)

    def set_camera_xml(self, geometry: ET) -> None:
        """
        Setup balise camera of xml.

        Args:
            geometry (ET): The xml.
        """
        intrinseque = ET.SubElement(geometry, "intrinseque")

        sensor = ET.SubElement(intrinseque, "sensor")
        ET.SubElement(sensor, "name").text = self.cam.name_camera

        calibration_date = ET.SubElement(sensor, "calibration_date")
        add_elem(calibration_date, B_TIME, [0, 0, 0, 0, 0, 0, None])

        ET.SubElement(sensor, "serial_number").text = format_xml("UNKNOWN")

        image_size = ET.SubElement(sensor, "image_size")
        add_elem(image_size, B_TCAM, [self.cam.width, self.cam.height])

        sensor_size = ET.SubElement(sensor, "sensor_size")
        add_elem(sensor_size, B_TCAM, [self.cam.width, self.cam.height])

        ppa = ET.SubElement(sensor, "ppa")
        add_elem(ppa, B_CAM, [int(self.cam.ppax), int(self.cam.ppay), int(self.cam.focal)])

        ET.SubElement(sensor, "pixel_size").text = str(0.000004)
